// main.cpp
// Gregory Rosenblatt

#include <fstream>
#include "virtualmachine.h"
#include "parsing.h"
#include "system.h"
using namespace MetaLang;
using namespace std;
using Test::System;

//#define LOAD_FROM_CONSOLE
#define LOAD_FROM_FILE

const int NUM_VARIABLES = 4;

int main()
{
    string codeList[num_codes] =
    {
        "output",
        "expose_state",
        "push_const",
        "push_var",
        "push_scope",
        "pop_scope",
        "assign",
        "store",
        "load",
        "add",
        "subtract",
        "multiply",
        "divide",
        "not",
        "and",
        "or",
        "equal",
        "not_equal",
        "greater",
        "greater_equal",
        "less",
        "less_equal",
        "jump",
        "jump_if_true",
        "jump_if_false",
        "call_func",
        "end_func",
        "pusharg_const",
        "pusharg_var",
        "call_command",
        "end"
    };
    
    typedef Callback<System, int, int>::Func SystemFunc;
    SystemFunc commandList[System::num_commands] =
    {
        &System::DrawStars,
        &System::CalcFactorial
    };
    
    System system;
    VirtualMachine<System> vm(system, commandList, System::num_commands);
    Scanner scanner(codeList, num_codes);
    vector<Instruction> instrList;
    string input;

#ifdef LOAD_FROM_CONSOLE

    cout << "Input script instructions:" << endl;
    while (input != "end")
    {
        getline(cin, input);
        if (scanner.ScanLine(input))
        {
            const CharArray& tokens = scanner.GetTokens();
            opcode code = static_cast<opcode>(tokens[0]);
            instrList.push_back(Instruction(code, &tokens[1], tokens.size()-1));
        }
        else
            cout << "Invalid Instruction!" << endl;
    }
    
    // as a safety precaution, it couldn't hurt to have redundant ends
    instrList.push_back(Instruction(op_end));

    // obtain a variable count
    size_t varCount;
    cout << "Input required variables count: ";
    cin >> varCount;
#endif

#ifdef LOAD_FROM_FILE
    // prompt user to enter a filename
    cout << "Enter path/name of script file: ";
    getline(cin, input);

    // attempt to open the file
    ifstream file(input.c_str());
    if (!file.good())
    {
        cout << "Could not open file!" << endl;
        return 0;
    }

    size_t lineNum = 1, varCount = NUM_VARIABLES;
    // feed data into scanner and build instructions
    while (!file.eof())
    {
	string line;
	getline(file, line);

        // scan the line
        if (scanner.ScanLine(line))
        {
            const CharArray& tokens = scanner.GetTokens();
            opcode code = static_cast<opcode>(tokens[0]);
            instrList.push_back(Instruction(code, &tokens[1], tokens.size()-1));
        }
        else
        {
            cout << "Invalid Instruction!" << endl;
            cout << "Line number: " << lineNum << endl;
            cout << "Line: " << line << endl;
        }               
        ++lineNum;
    }
    
    // as a safety precaution, it couldn't hurt to have redundant ends
    instrList.push_back(Instruction(op_end));
#endif

    Script script(instrList, varCount);
    
    // load the script and save the id
    size_t scriptID = vm.Load(script);

    // execute script by its id
    cout << endl << "***EXECUTING SCRIPT***" << endl;
    vm.Execute(scriptID);
    // check out the variable states
    cout << endl << "***VARIABLE STATES***" << endl;
    vm.ShowVariableState();

    // pause for input to prevent app from closing before getting a look
    getline(cin, input);
    return 0;
}
